import numpy as np
from scipy.stats import chi2
from pathlib import Path
import re
from nmrtoolbox.util import Range, ROI, AxisData, TableData, PeakOutsideOfMaskError, ParsePeakTableError
from nmrtoolbox.mask import Mask


class Peak:
    """
    A container for the values that describe a peak.
    """
    def __init__(self, prop_list, val_list, axis_list):
        """
        Populate a peak

        :param prop_list: parameter names (from VARS line)
        :param val_list: parameter values (line from peak table)
        :param axis_list: names of axes (e.g. 'X', 'Y', 'Z')
        """
        if not (len(prop_list) == len(val_list)):
            raise ValueError('properties and values must be lists of the same length')

        self.axis_list = axis_list

        # use the character in the format string to convert each value accordingly
        self.prop = dict()
        for p, v in zip(prop_list, val_list):
            # try to convert the value to int or float; then accept it as string
            try:
                self.prop[p] = int(v)
                continue
            except ValueError:
                pass
            try:
                self.prop[p] = float(v)
                continue
            except ValueError:
                pass
            self.prop[p] = v

    def get_par(self, par, axis_list='all'):
        """
        Get value of specified parameter for specified axes.

        :param par: nmrPipe peak table parameter name (e.g. *_PPM, HEIGHT)
                    The * is substituted with axis values for each label requested
        :param axis_list: list of axis identifiers to include (e.g. 'X' or default of 'all')
        :return: singleton value or list of values
        """
        if '*' not in par:
            par_set = [par]
        else:
            par_set = []

            # determine which labels to query
            if axis_list == 'all':
                axis_list = self.axis_list
            if isinstance(axis_list, str):
                axis_list = [axis_list]

            # perform substitutions of axis names into parameter name as needed
            for axis in axis_list:
                if axis not in self.axis_list:
                    raise KeyError(f'invalid axis: {axis}')
                par_set.append(par.replace('*', axis))

        out = []
        for par in par_set:
            try:
                out.append(self.prop[par])
            except KeyError:
                raise KeyError(f'property not found: {par}')

        if len(out) == 1:
            return out[0]
        else:
            return out

    def get_par_list(self):
        return list(self.prop.keys())

    def set_par(self, par, value):
        """
        Change or create a new parameter

        :param par: parameter name
        :param value: corresponding value
        :return:
        """
        self.prop[par] = value

    def print(self):
        """Print all the key/value pairs that define the peak"""
        for k, v in self.prop.items():
            print(f"{k:8s}: {v:>15}")

    def in_roi(self, roi: ROI):
        """
        Check if current peak is within the given ROI

        :param roi: ROI class object
        :return: bool
        """
        if set(roi.axis_list()) != set(self.axis_list):
            raise ValueError('the current peak and the given ROI do not have same set of axes')

        return all([roi.range(axis).contains(self.get_par(f'{axis}_PPM')) for axis in self.axis_list])

    def in_mask(self, mask: Mask, box_radius=2):
        """
        determine if the peak position (and some neighborhood around it) is in empty region of mask
        """
        try:
            return mask.ppm_in_mask(
                peak_ppm=self.get_par('*_PPM'),
                box_radius=box_radius,
            )
        except PeakOutsideOfMaskError as e:
            print(e)
            return False


class PeakTable:
    """
    A class for containing a collection of peaks (e.g. from a peak picker or synthetic peak generator)
    Sub-classes are defined for each peak table file format.
    """
    def __init__(self, file):
        """
        Populate a peak list.  Each type of peak list file is sub-classed from PeakTable.  The subclasses should have
        an __init__ that defines "pre" operations, calls super.__init__, defines "post" operations.  The __init__ here
        includes pre operations for ALL PeakTable subclasses and it includes a call to _parse_file (which should be
        defined within each subclass).  The subclass __init__ should then call _post to run operations that are defined
        for ALL PeakTable subclasses.  By crating an instance of a subclass, the following operations are performed:
            "pre" operations in subclass __init__
            "pre" operations in PeakTable __init__
            _parse_file (defined by subclass)
            "post" operations in PeakTable __init__
            "post" operations in subclass __init__
            _post (defined by PeakTable)

        Here is a template for subclass definition

        class SomeNewSubclass(PeakTable):
            def __init__(self, file):
                # pre
                <define instance variables used specifically by this subclass>

                # perform operations before the table file is parsed that are for ALL subclasses
                # call the _parse_file method that is defined by the subclass
                super().__init__(file)

                # post
                <perform operations after table file is parsed that are specific to this subclass>
                self._post() <- perform operations after the table file is parsed that are for ALL subclasses

        :param file: peak table file
        """
        self.file = Path(file)
        if not self.file.is_file():
            raise FileNotFoundError('peak table does not exist: {}'.format(self.file))

        # list of Peak objects
        self.peaks = []

        # metadata for each axis (e.g. X, Y, Z) - captured in an AxisData object:
        #   - labels (e.g.H1, N15)
        #   - num_pts
        #   - ppm_range (from ppm_high, ppm_low)
        self.axis_property = AxisData()

        # global properties that describe the collection of peaks
        self.table_property = TableData()

        # subclasses will each call _parse_file() to populate self.peaks and self.axis
        # subclasses can optionally include "pre" and "post" tasks in their __init__ to perform before/after parsing

        try:
            self._parse_file()
        except ValueError as e:
            raise ParsePeakTableError(f'failed to read peak table: {self.file}\n{e}')
        except:
            raise ParsePeakTableError(f'failed to read peak table: {self.file}')

    def _parse_file(self):
        """stub - subclasses must define this and populate self.peaks and self.axis"""
        raise NotImplementedError('sub-class must define a _parse_file() function')

    def _post(self):
        """operations to run after an instance of any subclass has been created"""
        try:
            # includes carrier frequency (i.e. PeakTablePipeInj)
            self.axis_property.get('carrier_frequency')
        except KeyError:
            try:
                # does NOT include carrier frequency (i.e. PeakTablePipeRec)
                # try to compute carrier frequency by using peak positions in ppm and Hz
                self._set_carrier_frequency_from_ppm_hz()
            except KeyError:
                print('not able to set carrier frequency automatically')
                pass
        try:
            self.get_par(par='*_HZ')
        except KeyError:
            try:
                self._set_peak_positions_hz()
            except KeyError:
                print('not able to compute peaks positions in Hz automatically')
                pass

    def _set_carrier_frequency_from_ppm_hz(self):
        """Determine carrier frequency from the ratio between peak positions in Hz and ppm"""
        # get average hz/ppm ratio across all peaks for each dimension
        # NOTE: peak tables from genSimTab do not have Hz values, so catch missing data
        try:
            ppm = self.get_par(par='*_PPM')
            hz = self.get_par(par='*_HZ')
            carrier = np.mean(hz / ppm, axis=0) * 1e6
        except KeyError as e:
            raise KeyError(e)

        for axis, cf in zip(self.axis_property.axis_list(), carrier):
            self.axis_property.set(
                axis=axis,
                name='carrier_frequency',
                value=cf,
                unit='Hz',
            )

    def _set_peak_positions_hz(self):
        """Use ppm peak positions and provided carrier frequencies to determine peak positions in Hz"""

        for axis in self.axis_list():
            try:
                peak_ppm = self.get_par(par=f'{axis}_PPM')
            except KeyError:
                raise KeyError(f'missing peak PPM data for axis: {axis}')
            try:
                cf = self.axis_property.get(
                    name='carrier_frequency',
                    axis=axis,
                )[0]
                if cf.unit == 'Hz':
                    cf_mhz = cf.value / 1e6
                elif cf.unit == 'MHz':
                    cf_mhz = cf.value
                else:
                    raise ValueError(f'do not recognize unit for carrier frequency: {cf.unit}')
                peak_hz = peak_ppm * cf_mhz
            except KeyError:
                raise KeyError(f'missing carrier frequency for axis: {axis}')
            for idx, peak in enumerate(self.peaks):
                peak.set_par(par=f'{axis}_HZ', value=peak_hz[idx])

    def num_peaks(self):
        """
        Count the peaks in a peak list

        :return: count
        """
        return len(self.peaks)

    def get_peak(self, index):
        """
        Get a peak, specified by index

        :param index: index within PeakTable (not necessarily same as "index" in the tab file)
        :return: PipePeak
        """
        try:
            return self.peaks[index]
        except IndexError:
            raise IndexError('asking for index = {:d}, max index = {:d}'.format(index, len(self.peaks)))

    def get_par(self, par, axis_list='all'):
        """
        Get all values for specified parameter for all peaks for specified axes.

        :param par: nmrPipe peak table parameter name (e.g. *_PPM, HEIGHT)
                    The * is substituted with axis values for each label requested
        :param axis_list: list of axis identifiers to include (e.g. 'X' or default of 'all')
        :return: numpy array
        """
        out = []
        for peak in self.peaks:
            out.append(peak.get_par(
                par=par,
                axis_list=axis_list,
            ))
        return np.array(out)

    def get_par_list(self):
        return self.peaks[0].get_par_list()

    def reduce(
            self,
            number=None,
            height=None,
            abs_height=None,
            roi=None,
            index=None,
            cluster_type=None,
            mask=None,
            box_radius=2,
            chi2prob=None,
    ):
        """
        reduce the peak list using filter criteria

        :param number: number of peaks to keep according to absolute value of peak HEIGHT
        :param height: keep all peaks with at least this HEIGHT
        :param abs_height: keep all peaks with absolute value at least this HEIGHT
        :param roi: keep peaks within the given ROI
        :param index: keep only peaks with given index values
        :param cluster_type: keep only peaks with given cluster type index
            NMRPipe currently uses: 1 = Peak, 2 = Random Noise, 3 = Truncation artifact
        :param mask: mask of True/False values on indel grid (e.g. use empirical peak table to set empty region)
        :param box_radius: defines size of box around peak position when querying it against empty mask
        :param chi2prob: remove peaks whose widths are outliers along any of the dimensions using chi2 probability
            expressed as p-value on [0,1] interval
        :return:
        """
        if number is not None:
            self.order_by_height()
            try:
                self.peaks = self.peaks[:number]
            except IndexError:
                # if the number of requested peaks is larger than number of actual peaks, just ignore and keep all
                pass

        elif height is not None:
            self.peaks = [p for p in self.peaks if p.get_par('HEIGHT') >= height]

        elif abs_height is not None:
            self.peaks = [p for p in self.peaks if abs(p.get_par('HEIGHT')) >= abs_height]

        elif roi is not None:
            if not isinstance(roi, ROI):
                raise TypeError('when reducing PeakTable by ROI, you must provide ROI class object')
            self.peaks = [p for p in self.peaks if p.in_roi(roi)]

        elif index is not None:
            if not isinstance(index, (list, tuple)):
                index = [index]
            try:
                self.peaks = [self.peaks[idx] for idx in index]
            except IndexError:
                raise IndexError('invalid index values for peak list')

        elif cluster_type is not None:
            if cluster_type not in [1, 2, 3]:
                print('NMRPipe peaks are categorized as:')
                print('  1 = Peak, 2 = Random Noise, 3 = Truncation artifact')
                raise TypeError('NMRPipe cluster_type must be one of the values above')
            self.peaks = [p for p in self.peaks if p.get_par('TYPE') == cluster_type]

        elif mask is not None:
            if not isinstance(mask, Mask):
                raise ValueError(f'provided mask should be of type nmrtoolbox.Mask (you provided {type(mask)}')
            self.peaks = [p for p in self.peaks if p.in_mask(mask=mask, box_radius=box_radius)]

        elif chi2prob is not None:
            idx_keep, idx_outlier = self.determine_outliers(chi2prob)
            self.reduce(index=idx_keep)

    def order_by_height(self):
        """Reorder the peak list by HEIGHT (positive down to negative)"""
        self.peaks = sorted(self.peaks, key=lambda x: x.prop['HEIGHT'], reverse=True)

    def axis_list(self):
        """Get the axis names"""
        return list(self.axis_property.axis_list())

    def label_list(self):
        """Get the axis labels"""
        return self.axis_property.get_value(name='label')

    def axis_label_list(self):
        """Get a list of tuples where each contains an axis and its label"""
        return [(ax, lab) for ax, lab in zip(self.axis_list(), self.label_list())]

    def determine_outliers(self, chi2prob):
        # get full width of each recovered peak in points along each of its dimensions
        w = self.get_par('*W')

        # Covariance matrix
        covariance = np.cov(w, rowvar=False)

        # Covariance matrix power of -1
        covariance_pm1 = np.linalg.matrix_power(covariance, -1)

        # Center point
        centerpoint = np.mean(w, axis=0)

        # the width of each peak along each dimension is compared to the mean width of all peaks along the dimension
        # the covariance matrix is used to normalize
        distances = (np.matmul(w - centerpoint, covariance_pm1) * (w - centerpoint)).sum(axis=1)

        # Cutoff (threshold) value from Chi-Square Distribution for detecting outliers
        cutoff = chi2.ppf(chi2prob, w.shape[1])

        # TODO
        #  stupid [0] needed to get ndarray out of tuple
        #  for now: put ndarray into list because PeakTable.reduce() seems to fail when index is provided as ndarray
        idx_keep = [x for x in np.where(distances <= cutoff)[0]]
        idx_outlier = [x for x in np.where(distances > cutoff)[0]]

        # return these lists so that other downstream analysis can be performed
        return idx_keep, idx_outlier


class PeakTablePipe(PeakTable):
    def __init__(self, file):
        """
        This class has subclasses for peak tables coming from INJECTED and RECOVERED peak tables; use those subclasses.
        You probably don't want to create a PeakTablePipe object - it is just a super-class to provide common tasks.
        """
        # pre
        self.VARS = None
        self.FORMAT = None

        super().__init__(file)

    def _parse_file(self):
        """
        Parser for peak tables from NMRPipe generated by peak picker (i.e. "REC") OR by genSimTab (i.e. "INJ").
        The PeakTablePipe class is subclassed to handle the differences in data presented in RECOVERED and INJECTED.
        """

        with open(self.file, 'r') as f_in:
            mode = None
            for line in f_in:
                line = line.strip()
                if line in ['', '#']:
                    continue

                if mode is None:
                    # recovered peak table starts with REMARK section
                    # injected peak table starts with genSimTab
                    if line.startswith('REMARK'):
                        mode = 'remark'
                    elif re.match('# [a-zA-Z_]*genSimTab.tcl', line):
                        # several variants of genSimTab.tcl are in use, so find all lines that might look like:
                        #   # genSimTab.tcl
                        #   # xzy_genSimTab.tcl
                        #   # new_genSimTab.tcl
                        mode = 'genSimTab'
                    else:
                        continue

                if mode == 'remark':
                    if line.startswith('REMARK'):
                        self.REMARK.append(line)
                        continue
                    else:
                        mode = 'data'

                if mode == 'genSimTab':
                    if line.startswith('#'):
                        self.genSimTab.append(line.strip().lstrip('#').rstrip('\\').strip())
                        continue
                    else:
                        mode = 'data'

                if mode == 'data':
                    if line.startswith('DATA'):
                        # axis properties are described like this
                        # example: DATA  X_AXIS HN           1   659   10.297ppm    5.798ppm
                        info = line.split()
                        try:
                            if info[1] == 'CLUSTER':
                                # some filtered peak tables include CLUSTER info - ignore these lines
                                # example: DATA  CLUSTER X_AXIS +/- 7
                                continue
                            elif info[1].endswith('_AXIS'):
                                axis = info[1].split('_')[0]
                                self.axis_property.set(
                                    axis=axis,
                                    name='label',
                                    value=info[2],
                                )
                                self.axis_property.set(
                                    axis=axis,
                                    name='num_pts',
                                    value=int(info[4]),
                                    unit='count'
                                )
                                self.axis_property.set(
                                    axis=axis,
                                    name='range',
                                    value=Range(
                                            min=float(info[5].rstrip('pm')),
                                            max=float(info[6].rstrip('pm')),
                                        ),
                                    unit='ppm',
                                )
                                continue
                            else:
                                raise ValueError(f'unrecognized line in peak table: {line}')
                        except (IndexError, ValueError):
                            raise ValueError('metadata missing from peak table header')
                    else:
                        mode = 'vars'

                if mode == 'vars':
                    if line.startswith('VARS'):
                        self.VARS = line.split()[1:]
                        # this is a single line of data, so get it and move on to next mode
                        mode = 'format'
                        continue

                if mode == 'format':
                    if line.startswith('FORMAT'):
                        self.FORMAT = line.split()[1:]
                        # this is a single line of data, so get it and move on to next mode
                        mode = 'peak'
                        continue

                if mode == 'peak':
                    if line.startswith('#'):
                        # comments are mixed into the synthetic peak lists - just ignore them
                        continue
                    else:
                        self.peaks.append(Peak(
                            prop_list=self.VARS,
                            val_list=line.split(),
                            axis_list=self.axis_list(),
                        ))


class PeakTablePipeRec(PeakTablePipe):
    """This subclass is for capturing the peaks "recovered" from an experiment by the NMRPipe peak picker"""
    def __init__(self, file):
        # pre
        self.REMARK = []

        try:
            super().__init__(file)
        except:
            raise ParsePeakTableError(f'failed to read peak table as type: PeakTablePipeRec\n{file}')

        # post
        self._get_noise()
        self._get_category_count()
        self._get_LW()

        self._post()

    def _get_LW(self):
        try:
            LW_list = self.get_par('*W').mean(axis=0)
        except KeyError as e:
            raise KeyError(e)

        for axis, LW in zip(self.axis_property.axis_list(), LW_list):
            self.axis_property.set(
                axis=axis,
                name='LW',
                value=LW,
                unit='pts',
            )

    def _get_noise(self):
        for line in self.REMARK:
            if line.startswith('REMARK Noise:'):
                # example: "REMARK Noise: 91905.9, Chi2-Threshold: 1.000000e-04, Local Adjustment: None"
                self.table_property.set(
                    name='noise',
                    value=float(line.split()[2].strip(',')),
                    unit='au',
                )

    def _get_category_count(self):
        for line in self.REMARK:
            if line.startswith('REMARK Total Peaks:'):
                # example: "REMARK Total Peaks: 60425, Good Peaks: 40513, Questionable Peaks: 19912"
                self.table_property.set(
                    name='peak_count_total',
                    value=int(line.split()[3].strip(',')),
                    unit='count'
                )
                self.table_property.set(
                    name='peak_count_good',
                    value=int(line.split()[6].strip(',')),
                    unit='count',
                )
                self.table_property.set(
                    name='peak_count_questionable',
                    value=int(line.split()[9].strip(',')),
                    unit='count',
                )


class PeakTablePipeInj(PeakTablePipe):
    """This subclass is for capturing the synthetic peaks "injected" by the NMRPipe genSimTab.tcl tool"""
    def __init__(self, file, carrier_frequency):
        # pre
        self.genSimTab = []
        self.carrier_frequency = carrier_frequency

        try:
            super().__init__(file)
        except:
            raise ParsePeakTableError(f'failed to read peak table as type: PeakTablePipeInj\n{file}')

        # post
        self.genSimTab = ' '.join(self.genSimTab)
        self._set_maxLW()
        self._set_carrier_frequency()
        self._post()

    def _set_carrier_frequency(self):
        # the injected peaks are presumed to be inserted into an empirical data set, so read the
        # carrier frequency out of the empirical peak table, because it is not included in the
        # injected peak table
        # carrier_frequency = self.empirical_table.axis_property.get(name='carrier_frequency')

        # TODO: not great - this is using an implicitly ordered list to set values
        for ax, cf in zip(self.axis_list(), self.carrier_frequency):
            self.axis_property.set(
                axis=ax,
                name='carrier_frequency',
                value=cf.value,
                unit=cf.unit,
            )

    def _set_maxLW(self):
        # parse through the tokens to find maxLW values for each axis
        tokens = self.genSimTab.split()
        for axis in self.axis_list():
            try:
                idx = tokens.index(f'-{axis.lower()}wMax')
                self.axis_property.set(
                    axis=axis,
                    name='maxLW',
                    value=float(tokens[idx + 1]),
                    unit='Hz',
                )
            except (ValueError, IndexError):
                pass



#!/usr/bin/env python3
import time

import copy
import os
import sys
import argparse
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path
from scipy.spatial import distance_matrix
from itertools import combinations

from nmrtoolbox.peak import PeakTable, PeakTablePipeRec, PeakTablePipeInj
from nmrtoolbox.util import ROI, ParsePeakTableError, pairwise_weighted_norm
from nmrtoolbox.mask import Mask


class roc:
    """
    A class for performing receiver operator characteristic (ROC) analysis on NMR peak lists
    """

    def __init__(
            self,
            recovered_table,
            injected_table,
            lw_scalar=1,
            number=None,
            height=None,
            abs_height=None,
            roi_list: list = None,
            index=None,
            cluster_type=None,
            mask_file=None,
            box_radius=2,
            chi2prob=None,
    ):
        """
        :param recovered_table: table (file) of recovered peaks from NMRPipe->pkDetect3D.tcl
        :param injected_table: table (file) of synthetic peaks from NMRPipe->genSimTab.tcl
        :param lw_scalar: scalar of an injected peak's linewidth to define a region in which a recovered peak is valid
        :param number: number of peaks to keep according to absolute value of peak HEIGHT
        :param height: keep all peaks with at least this HEIGHT
        :param abs_height: keep all peaks with absolute value at least this HEIGHT
        :param roi_list: region of interest to filter the recovered_peaklist
            by in order to assess for corresponding synthetic peaks,
            format is a list of [x_lower_bound_ppm, x_upper_bound_ppm, y_lower_bound_ppm, y_upper_bound_ppm, ...]
        :param index: keep only peaks with given index values
        :param cluster_type: keep only peaks with given cluster type index
            NMRPipe currently uses: 1 = Peak, 2 = Random Noise, 3 = Truncation artifact
        :param mask_file: spectra of the empty region used to filter out empirical from synthetic injected spectra that
            ROC is looking to recover; an alternative to providing an already filtered peaklist
        :param box_radius: defines size of box around peak position when querying it against empty mask
        :param chi2prob: remove peaks whose widths are outliers along any of the dimensions using chi2 probability
            expressed as p-value on [0,1] interval
        """
        # This ROC class accepts filtering criteria as a convenience to users so they do not have to read in peak tables
        # and call filtering functions before running roc.  So process the inputs for the peak table files and the roi
        # and then call the peak.reduce command to handle the rest.

        # ================================================
        if isinstance(recovered_table, PeakTable):
            self.recovered_peaks = recovered_table
        else:
            try:
                self.recovered_peaks = PeakTablePipeRec(
                    file=recovered_table,
                )
            except FileNotFoundError as e:
                raise FileNotFoundError(e)
            except ParsePeakTableError as e:
                raise ParsePeakTableError(e)

        # ================================================
        if isinstance(injected_table, PeakTable):
            self.injected_peaks = injected_table
        else:
            try:
                self.injected_peaks = PeakTablePipeInj(
                    file=injected_table,
                    carrier_frequency=self.recovered_peaks.axis_property.get('carrier_frequency'))
            except FileNotFoundError as e:
                raise FileNotFoundError(e)
            except ParsePeakTableError as e:
                raise ParsePeakTableError(e)

        # ================================================
        # validate some aspects of the peak table metadata and then move on to filtering operations
        self._validate_input()

        # ================================================
        self.roi = None
        if roi_list is not None:
            axis_list = self.injected_peaks.axis_list()
            if len(roi_list) != 2*len(axis_list):
                raise ValueError('there should be a min and max value in ppm for each axis')

            range_list = []
            while roi_list:
                range_list.append([roi_list.pop(0), roi_list.pop(0)])

            self.roi = ROI(
                axis_list=axis_list,
                range_list=range_list,
            )

        # ================================================
        self.mask = None
        if mask_file is not None:
            print('Reading mask file.  This may take a while...')

            self.mask_file = mask_file
            self.mask = Mask(file=self.mask_file)

        # ================================================
        # Get the max LW values used by genSimTab and then expand by a lw_scalar to define the distance along each axis
        # used to define if a recovered peak is within the accepted range of an injected peak to be considered a match.
        self.lw_scalar = lw_scalar
        self.peak_range_hz = np.array(self.injected_peaks.axis_property.get_value('maxLW') * self.lw_scalar)

        # ================================================
        # run the reduce commands (if any parameters are not given to roc, then they are None and reduce skips them)
        self.recovered_peaks.reduce(number=number)
        self.recovered_peaks.reduce(height=height)
        self.recovered_peaks.reduce(abs_height=abs_height)
        self.recovered_peaks.reduce(roi=self.roi)
        self.recovered_peaks.reduce(index=index)
        self.recovered_peaks.reduce(cluster_type=cluster_type)
        self.recovered_peaks.reduce(
            mask=self.mask,
            box_radius=box_radius,
        )

        # ================================================
        if chi2prob is not None:
            # could do this directly with: self.recovered_peaks.reduce(chi2prob=chi2prob)
            # BUT: the commands of reduce are reproduced here so that intermediate data can be captured
            # and used for outlier plotting

            # save instances of ALL peaks before removing the outlier peaks
            self.recovered_peaks_all = copy.deepcopy(self.recovered_peaks)
            self.recovered_peaks_outliers = copy.deepcopy(self.recovered_peaks)

            # partition all peaks into outliers and not outliers
            idx_keep, idx_outlier = self.recovered_peaks.determine_outliers(chi2prob=chi2prob)
            self.recovered_peaks_outliers.reduce(index=idx_outlier)
            self.recovered_peaks.reduce(index=idx_keep)

        # ================================================
        # create instance variables to capture the output from running ROC and then run ROC
        # initialize the RR and FDR lists to include the origin
        self.RR = [0]
        self.FDR = [0]
        self.recovered_peaks_true_ppm = []
        self.recovered_peaks_false_ppm = []
        self._recovery_rate()

        # compute metrics from ROC data
        self.AUC = self._AUC()
        self.DPC, self.DPC_index = self._DPC()
        self.MRMF = self._MRMF()

    def _validate_input(self):
        # sometimes proton is labeled as 1H or HN, so standardize on the more common HN
        if '1H' in self.recovered_peaks.label_list() and 'HN' in self.injected_peaks.label_list():
            for axis, label in self.recovered_peaks.axis_label_list():
                if label == '1H':
                    self.recovered_peaks.axis[axis].label = 'HN'

        if set(self.injected_peaks.label_list()) != set(self.recovered_peaks.label_list()):
            raise ValueError('The injected and recovered peak tables have different axis labels.')

        if hasattr(self, 'roi'):
            if len(self.injected_peaks.axis_list()) != len(self.roi.axis_list()):
                raise ValueError(
                    'Number of dimensions for injected and recovered peak lists are not in agreement with ROI list')

        if self.injected_peaks.num_peaks() == 0:
            raise ValueError('No recorded peaks in injected peak table')

        if self.recovered_peaks.num_peaks() == 0:
            raise ValueError('No peaks recovered by peak picker')

    def _recovery_rate(self):
        # ordering of peaks matters
        # sort the recovered peaks by height
        #   => the most intense recovered peaks have priority in being matched to closest injected peaks
        self.recovered_peaks.order_by_height()

        injected_peaks_hz = self.injected_peaks.get_par(par="*_HZ")
        recovered_peaks_hz = self.recovered_peaks.get_par(par="*_HZ")
        recovered_peaks_ppm = self.recovered_peaks.get_par(par="*_PPM")

        # compute pairwise distances from between recovered and injected peaks - to be used for defining matches
        # TODO:
        #  Currently finding closest injected peak to each recovered peak with unweighted L2, but at least it is in
        #  Hz and not ppm.  Eventually, the closest match identified here is more accurately assessed below by taking
        #  the distance along along each axis and comparing to a cutoff distance defined as multiple of linewidth.
        #  So... should a weighted L2 be sued here?
        # start_time = time.time()
        pairwise_dist = distance_matrix(recovered_peaks_hz, injected_peaks_hz)
        # print(f"scipy time: {time.time() - start_time}")

        # start_time = time.time()
        # pairwise_dist2 = pairwise_weighted_norm(recovered_peaks_hz, injected_peaks_hz, w=np.array((1, 1, 1)))
        # print(f"custom L2 time: {time.time() - start_time}")

        # for each recovered peak find the closest injected peak
        idx_recovered2injected = np.argmin(pairwise_dist, axis=1)

        # running tallies used to compute the recovery rate and false discovery rate
        count_true = 0
        count_false = 0

        # start with all injected peaks and remove them as they are matched to a recovered peak
        idx_injected_peak_available = set(range(self.injected_peaks.num_peaks()))

        for idx_recovered_peak, idx_injected_peak in enumerate(idx_recovered2injected):
            # distance along each axis from injected to nearest recovered peak
            distance_hz = np.abs(recovered_peaks_hz[idx_recovered_peak] - injected_peaks_hz[idx_injected_peak])

            # check if the closest injected peak is still available AND if it's inside the allowable neighborhood
            if (idx_injected_peak in idx_injected_peak_available) and (distance_hz <= self.peak_range_hz).all():
                count_true += 1
                self.recovered_peaks_true_ppm.append(recovered_peaks_ppm[idx_recovered_peak])
                idx_injected_peak_available.remove(idx_injected_peak)
            else:
                count_false += 1
                self.recovered_peaks_false_ppm.append(recovered_peaks_ppm[idx_recovered_peak])

            self.FDR.append(count_false / (count_false+count_true))
            self.RR.append(count_true / self.injected_peaks.num_peaks())

        # Appending the top right corner (extending the "plateau" of the ROC curve to reach the right edge)
        self.FDR.append(1)
        self.RR.append(max(self.RR))

        # Appending the bottom right corner
        self.FDR.append(1)
        self.RR.append(0)

    def _DPC(self):
        corner = np.array([[0, 1]])
        roc = np.column_stack((self.FDR, self.RR))
        dst = distance_matrix(corner, roc)

        dpc = np.min(dst)
        dpc_index = np.argmin(dst)

        return dpc, dpc_index

    def _AUC(self):
        """Area under the curve"""
        if len(self.RR) == len(self.FDR):
            n = len(self.RR)  # number of corners
        else:
            raise ValueError('RR and FDR are not the same length')

        # reference: https://math.stackexchange.com/questions/492407/area-of-an-irregular-polygon
        area = 0.0
        for i in range(n):
            j = (i + 1) % n
            area += self.FDR[i] * self.RR[j]
            area -= self.FDR[j] * self.RR[i]
        area = abs(area) / 2.0

        return area

    def _MRMF(self):
        """Maximum Recovery at Minimum False"""
        # construct an iterator that finds the first false positive and returns the previous index
        # use next() to only evaluate the iterator until the first result is found
        # provide len(self.FDR) as default in case no false positive is found
        idx = next((i-1 for i, v in enumerate(self.FDR) if v > 0), len(self.FDR))

        # if the very first recovered peak is false, that would result in an invalid idx=-1 (catch that case)
        # otherwise return the recovery rate at the index
        if idx >= 0:
            return self.RR[idx]
        else:
            return 0

    def print_stats(self):
        print(f"AUC: {self.AUC:5.4f}")
        print(f"DPC: {self.DPC:5.4f}")
        print(f"DPC_index: {self.DPC_index:d}")
        print(f"MRMF: {self.MRMF:5.4f}")

    def plot_outliers(self, file_out='outliers.pdf', show_figure=False):
        """Plot histograms of peak widths for all recovered peaks and for outliers vs keepers"""

        try:
            peak_width_all = self.recovered_peaks_all.get_par(par="*W")
            peak_width_keep = self.recovered_peaks.get_par(par="*W")
            peak_width_out = self.recovered_peaks_outliers.get_par(par="*W")
        except AttributeError:
            print('can only plot outlier data if you use chi2prob outlier filtering')
            return

        fig, axs = plt.subplots(3, 3, sharex='col', sharey='col')
        plt.suptitle('Histograms of recovered peak widths')
        # iterate through dimension indices (columns of the figure)
        for dim in [0, 1, 2]:
            # row 0: peak widths for ALL peaks on dimension
            ax = axs[0, dim]
            n, bins, patches = ax.hist([p[dim] for p in peak_width_all], bins=20)
            ax.set_xlim(.9*np.min(bins), 1.1*np.max(bins))
            ax.set_ylim(-0.1*np.max(n), 1.1*np.max(n))
            ax.set_title(self.recovered_peaks.label_list()[dim])

            # row 1: peak widths for KEEPERS on dimension
            ax = axs[1, dim]
            ax.hist([p[dim] for p in peak_width_keep], bins=bins)

            # row 2: peak widths for OUTLIERS on dimension
            ax = axs[2, dim]
            ax.hist([p[dim] for p in peak_width_out], bins=bins)

        for a in axs[-1, :]:
            a.set_xlabel('peak width\n[points]')
        axs[0, 0].set_ylabel('ALL\n[count]')
        axs[1, 0].set_ylabel('KEEPER\n[count]')
        axs[2, 0].set_ylabel('OUTLIER\n[count]')
        fig.tight_layout()

        Path(file_out).parent.absolute().mkdir(exist_ok=True)
        fig.savefig(file_out, format='pdf', dpi=1200)
        if show_figure:
            plt.show()
        plt.close()

    def plot_roc(self, file_out='ROC.pdf', show_figure=False):

        fig, ax = plt.subplots()
        ax.set_aspect('equal')
        plt.xlabel('False Discovery Rate', fontsize=14)
        plt.ylabel('Recovery Rate', fontsize=14)
        plt.plot(self.FDR[:-1], self.RR[:-1],
                 marker='o', mew=0, mfc='blue', ms=6,   # marker properties
                 c='.7', lw=2,                          # line properties
                 label='ROC curve',
                 zorder=1)
        plt.xlim(-0.02, 1.02)
        plt.ylim(-0.02, 1.02)

        # put a big red dot on the point that is closest to perfect classifier
        plt.scatter(self.FDR[self.DPC_index], self.RR[self.DPC_index],
                    marker='s', c='red', s=30,
                    label='DPC reference',
                    zorder=2)

        plt.legend(loc='lower right')

        Path(file_out).parent.absolute().mkdir(exist_ok=True)
        fig.savefig(file_out, format='pdf', dpi=1200)
        if show_figure:
            plt.show()
        plt.close()

    def plot_peaks(self, dir_out=os.getcwd(), file_basename='peaks', show_figure=False):

        # TODO: this axis and dimension handling is very confusing
        num_dim = len(self.injected_peaks.axis_list())
        temp_slice_axes = combinations(list(range(1, num_dim + 1)), 2)
        slice_axes = [i for i in temp_slice_axes]

        injected_peaks_ppm = self.injected_peaks.get_par(par="*_PPM")

        for dims in slice_axes:
            xdim = int(dims[0]) - 1
            ydim = int(dims[1]) - 1

            inj_xdim_ppm = [p[xdim] for p in injected_peaks_ppm]
            inj_ydim_ppm = [p[ydim] for p in injected_peaks_ppm]

            trueRecov_xdim_ppm = [p[xdim] for p in self.recovered_peaks_true_ppm]
            trueRecov_ydim_ppm = [p[ydim] for p in self.recovered_peaks_true_ppm]

            falseRecov_xdim_ppm = [p[xdim] for p in self.recovered_peaks_false_ppm]
            falseRecov_ydim_ppm = [p[ydim] for p in self.recovered_peaks_false_ppm]

            fig, ax = plt.subplots()
            fig.subplots_adjust(right=0.75)
            plt.xlabel(f"{self.recovered_peaks.label_list()[xdim]} Domain [ppm]")
            plt.ylabel(f"{self.recovered_peaks.label_list()[ydim]} Domain [ppm]")
            plt.scatter(trueRecov_xdim_ppm, trueRecov_ydim_ppm,
                        c='#C5C9C7', marker='s', label='True Recovered')
            plt.scatter(falseRecov_xdim_ppm, falseRecov_ydim_ppm,
                        c='k', marker='x', label='Falsely Recovered')
            plt.scatter(inj_xdim_ppm, inj_ydim_ppm,
                        c='g', marker='v', label='injected')
            plt.legend(loc=(1.02, 0.15), prop={'size': 8})
            plt.title("Categorization of recovered peaks relative to injected peaks",
                         fontsize=10)

            filename = f"{file_basename}_{xdim + 1}_{ydim + 1}.pdf"
            file_out = Path(dir_out) / filename
            Path(file_out).parent.absolute().mkdir(exist_ok=True)

            plt.savefig(file_out, format='pdf', dpi=1200)
            if show_figure:
                plt.show()
            plt.close()


def parse_args():
    parser = argparse.ArgumentParser(description='You can add a description here')
    parser.add_argument('--recovered_table', required=True)
    parser.add_argument('--injected_table', required=True)
    parser.add_argument('--lw_scalar', type=float,
                        help='scalar multiple of linewidth for injected peak to define valid region for recovered peak')
    parser.add_argument('--number', type=int,
                        help='number of peaks to keep (starting from most intense)')
    parser.add_argument('--height', type=float,
                        help='')
    parser.add_argument('--abs_height', type=float,
                        help='')
    parser.add_argument('--roi_list', nargs='*', type=float,
                        help='min and max values in ppm for each dimension (min1 max1 min2 max2 ...)')
    parser.add_argument('--index', nargs='*', type=int,
                        help='peak indices to keep (indexing by position in peak list, NOT by indexing embedded in peak file)')
    parser.add_argument('--cluster_type', choices=[1, 2, 3],
                        help='keep only peaks with given cluster type index. NMRPipe currently uses: 1 = Peak, 2 = Random Noise, 3 = Truncation artifact')
    parser.add_argument('--mask_file', type=str,
                        help='path to file containing the mask of the empty region (tabular format from ConnjurST')
    parser.add_argument('--box_radius', type=int,
                        help='size of box around peak position when querying it against empty mask')
    parser.add_argument('--chi2prob', type=float,
                        help='chi square probability [0-1] used to remove peaks with outlier width')
    parser.add_argument('--print_stats', action='store_true',
                        help='print all ROC metric values to stdout')
    parser.add_argument('--plot_roc', action='store_true',
                        help='generate the roc plot and save to file')
    parser.add_argument('--plot_outliers', action='store_true',
                        help='generate histogram of outlier peaks if chi2prob is used for outlier removal')
    parser.add_argument('--plot_peaks', action='store_true',
                        help='generate projections of injected and recovered peak positions and save to file')
    return parser.parse_args()


def main():
    # parse the arguments from command line input and execute the nuscon workflow
    args = parse_args()

    try:
        my_roc = roc(
            recovered_table=args.recovered_table,
            injected_table=args.injected_table,
            lw_scalar=args.lw_scalar,
            number=args.number,
            height=args.height,
            abs_height=args.abs_height,
            roi_list=args.roi_list,
            index=args.index,
            cluster_type=args.cluster_type,
            mask_file=args.mask_file,
            box_radius=args.box_radius,
            chi2prob=args.chi2prob,
        )
        if args.print_stats:
            my_roc.print_stats()
        if args.plot_roc:
            my_roc.plot_roc()
        if args.plot_outlier:
            my_roc.plot_outliers()
        if args.plot_peaks:
            my_roc.plot_peaks()

    except (SystemExit, EnvironmentError, OSError) as e:
        print(e)
        sys.exit()



from dataclasses import dataclass
from collections import defaultdict
import numpy as np

class PeakOutsideOfMaskError(Exception):
    # use this when the position of a Peak is outside of the ppm ranges covered by a Mask
    pass


class ParsePeakTableError(Exception):
    pass


@dataclass
class Range:
    """
    Store a min and max value to define a range
    """
    min: float
    max: float

    def contains(self, val):
        """
        Check if value is in range
        :param val: value to check (float or int)
        :return: bool
        """
        if not isinstance(val, (float, int)):
            raise TypeError('must enter a float or int')
        return self.min <= val <= self.max


class Data:
    def __init__(self, name, value, unit):
        # NOT using @dataclass because it requires explicit typing; want to allow value to be int, float, str, etc.
        self.name = name
        self.value = value
        self.unit = unit

    def print(self):
        print(f"{self.name} = {str(self.value)} [{self.unit}]")


class TableData:
    def __init__(self):
        self.data = {}

    def set(self, name, value, unit=None):
        self.data[name] = Data(name=name, value=value, unit=unit)

    def get(self, name):
        try:
            return self.data[name]
        except KeyError as e:
            raise KeyError(e)

    def print(self):
        [d.print() for d in self.data.values()]


class AxisData:
    def __init__(self):
        """dictionary of axis names where the values are dictionaries of Data objects"""
        self.data = defaultdict(dict)

    def axis_list(self):
        return list(self.data.keys())

    def set(self, axis, name, value, unit=None):
        axis = axis.upper()
        self.data[axis][name] = Data(name=name, value=value, unit=unit)

    def get(self, name, axis=None, data_field=None):
        """
        Get Data object for given name across the axis list.
        axis: name of axis (str) or list of axis names
        data_field: extract that field from Data objects or return entire Data object input not given
        """
        if axis is None:
            ax_list = self.axis_list()
        else:
            if not isinstance(axis, list):
                axis = [axis]
            try:
                ax_list = [ax.upper() for ax in axis]
            except AttributeError:
                raise AttributeError(f"axis should probably be 'X', 'Y', 'Z' or similar. unable to use: {str(axis)}")

        out = []
        try:
            for ax in ax_list:
                if data_field is None:
                    # return the Data object
                    out.append(self.data[ax][name])
                else:
                    # return the value of one field in the Data object
                    out.append(getattr(self.data[ax][name], data_field))
        except KeyError as e:
            raise KeyError(e)
        return out

    def get_value(self, name, axis=None):
        return self.get(axis=axis, name=name, data_field='value')

    def print(self):
        for axis in self.axis_list():
            print(f"axis = {axis}")
            [d.print() for d in self.data[axis].values()]


class ROI:
    def __init__(self, axis_list, range_list):
        """
        Define a region of interest (ROI) by giving the min/max values along each axis
        :param axis_list: axis names (e.g. use 'X', 'Y', 'Z' as used by PeakTable)
        :param range_list: list or tuple containing a min/max list or tuple for each axis
        """
        # if there is only one axis, then convert singleton inputs into lists
        if not isinstance(axis_list, (list, tuple)):
            axis_list = [axis_list]
        if not isinstance(range_list[0], (list, tuple)):
            range_list = [range_list]
        if not all(len(r) == 2 for r in range_list):
            raise ValueError('must enter a min/max ppm value for each axis as a list of lists')

        if len(axis_list) != len(range_list):
            raise ValueError('must enter a min/max ppm value for each axis')

        # dictionary to contain range values for each axis
        # it is best to access this data through self.range() which can allow no inputs (i.e. get the entire _range
        # dict) or take an input to specify a single axis (i.e. get the Range object for that axis)
        self._range = dict()
        for axis, vals in zip(axis_list, range_list):
            if not isinstance(vals, (list, tuple)):
                raise TypeError(f'must provide the min/max values as a list or tuple for axis: {axis}')
            if len(vals) != 2:
                raise ValueError(f'must provide min and max as a tuple or list of 2 values for axis: {axis}')
            try:
                # make sure the "min" field gets the minimum value and "max" field gets maximum value
                # (regardless of order given).   NMR people are strange in how they orient these values
                self._range[axis] = Range(
                    min=float(min(vals)),
                    max=float(max(vals)),
                )
            except ValueError:
                raise ValueError(f'could not convert ROI range values into floats: {range_list}')

    def axis_list(self):
        return self._range.keys()

    def range(self, axis=None):
        if axis is None:
            # default => return entire _range dictionary
            return self._range
        else:
            try:
                return self._range[axis]
            except KeyError:
                raise KeyError(f'axis not found: {axis}')
            except TypeError:
                raise TypeError('axis should be a string')


def weightedL2(a, b, w):
    """weighted norm of vectors a and b using weighting coefficients from w"""
    q = a - b
    return np.sqrt((w * q * q).sum())


def pairwise_weighted_norm(A, B, w=None):
    """
    A and B are matrices where each row is a vector of n dimensions.
    w is a vector of length n containing weighting factors for an L2 norm
    """
    if not (isinstance(A, np.ndarray) and isinstance(B, np.ndarray)):
        raise TypeError('must provide inputs as numpy.array')
    rowA, colA = A.shape
    rowB, colB = B.shape
    if colA != colB:
        raise ValueError('A and B must have same number dimensions in each column vector')

    if w is not None:
        try:
            w = np.asarray(w)
        except:
            raise TypeError('weights must be provided as data type that can be converted to numpy.array')
    else:
        w = np.ones(colA)
    numW = len(w)

    if numW != colA:
        raise ValueError('w must have the same number of weighting factors as A and B have dimensions')

    N = np.zeros((rowA, rowB))
    for indexA, a in enumerate(A):
        for indexB, b in enumerate(B):
            N[indexA][indexB] = weightedL2(a, b, w)
    return N


def pairwise_weighted_norm2(A, B, w=None):
    """
    A and B are matrices where each row is a vector of n dimensions.
    w is a vector of length n containing weighting factors for an L2 norm
    """
    if not (isinstance(A, np.ndarray) and isinstance(B, np.ndarray)):
        raise TypeError('must provide inputs as numpy.array')
    rowA, colA = A.shape
    rowB, colB = B.shape
    if colA != colB:
        raise ValueError('A and B must have same number dimensions in each column vector')

    if w is not None:
        try:
            w = np.asarray(w)
        except:
            raise TypeError('weights must be provided as data type that can be converted to numpy.array')
    else:
        w = np.ones(colA)
    numW = len(w)

    if numW != colA:
        raise ValueError('w must have the same number of weighting factors as A and B have dimensions')

    N = np.zeros((rowA, rowB))
    for indexA, a in enumerate(A):

        Q = B - a
        N[indexA] = np.sqrt((w * np.matmul(Q, Q.T)).sum())

    return N


import numpy as np
from pathlib import Path
import xml.etree.ElementTree as ET
import itertools
from nmrtoolbox.util import PeakOutsideOfMaskError


class Mask:
    """
    A class for defining a subset of the Nyquist grid.  The intended use is for NUScon where synthetic peaks are
    injected into the empty regions of an empirical spectrum.  This mask can define the empty regions so that the
    recovered peak list can be restricted to just regions where the synthetic peaks have been injected.
    The Mask is a numpy array sized to match the Nyquist grid with bool values where True indicates empty.
    """

    def __init__(self, file):
        """
        :param file: file containing the empty regions of an empirical spectrum, as generated by CONNJUR
        """
        try:
            self.file = Path(file)
        except TypeError:
            raise TypeError(f'provide path to mask file.  input path not recognized: {file}')
        if not self.file.is_file():
            raise FileNotFoundError(f'did not find mask file: {self.file}')

        # the xml includes custom namespace prefixes which need to be defined and used by xml.find()
        self.ns = {
            'connjur': "https://raw.githubusercontent.com/CONNJUR/CONNJUR_ML/master/connjur_ml.xsd",
            'builder': "https://raw.githubusercontent.com/CONNJUR/CONNJUR_ML/master/connjur_ml.xsd",
            'xsi': "http://www.w3.org/2001/XMLSchema-instance",
            'premis': "http://www.loc.gov/premis/v3",
        }

        # parse the tabular data file - capture the sections of the file and several properties
        self.header = []
        self.xml = None
        self.mask = None
        self.ndims = None
        self.points = None
        self._parse_file()

        if self.ndims != len(self.points):
            raise ValueError(f'number of dims does not match the number of axes defines')

        # derive info about the range of values covered by the mask and the ppm spacing of the grid points
        # this is used to map arbitrary peak positions (in ppm) onto the mask grid (points)
        self.first_pt_ppm = None
        self.last_pt_ppm = None
        self.pts_per_ppm = None
        self._get_axis_properties()

    def _parse_file(self):
        """parse tabular file that CWB generates - it contains a header, xml metadata, and tabular data"""

        xml_list = []

        with open(self.file) as file:
            mode = 'header'
            while line := file.readline().rstrip():
                # -------------------------------------------
                # 1. read the plain text header
                if mode == 'header':
                    if line.startswith('<?xml'):
                        # found the xml - move onto the next parsing stage with this line
                        mode = 'xml'
                    else:
                        # reading the plain text header
                        self.header.append(line)
                        continue
                # -------------------------------------------
                # 2. read the xml metadata
                if mode == 'xml':
                    # capture the line
                    xml_list.append(line)

                    if '</premis:premis>' in line:
                        # found the end of the xml header => parse it, init the mask, and move on
                        self.xml = ET.fromstringlist(xml_list)
                        self.ndims = self.get_xml('connjur:spectralAxes', dtype=int)[0]
                        self.points = self.get_xml('connjur:totalPoints', dtype=int)
                        self.mask = np.ones(self.points, dtype=bool)

                        # move onto the next parsing stage
                        mode = 'predata'

                    continue
                # -------------------------------------------
                # 3. skip anything that might appear after xml and before tabular data
                if mode == 'predata':
                    if "tabular data follows" in line:
                        # move onto next parsing stage
                        mode = 'table_header'
                        continue
                # -------------------------------------------
                # 4. skip table header
                if mode == 'table_header':
                    mode = 'data'
                    continue
                # -------------------------------------------
                # 5. read the tabular data
                # build a mask from the tabular data
                if mode == 'data':
                    vals = line.split()
                    # NOTE: since this is a mask the values are expected to be 0.0000 (not empty) or 1.0000 (empty)
                    # need float() to parse the string and then convert to bool()
                    # NOTE: mask is initialized as True, so just find and record the False values
                    # (there should be far fewer non-empty values (False) in the mask)
                    if not bool(float(vals[-1])):
                        indel = tuple(map(int, vals[:self.ndims]))
                        # ignoring phase for now
                        # phase = vals[-2]
                        self.mask[indel] = False
                    continue

    def _get_axis_properties(self):
        """
        Use metadata in xml of the mask file to compute how many points there are per ppm along each axis.  This will be
        used to convert an arbitrary peak's position (in ppm) into its corresponding location in the mask (in pts).
        """

        # the mask files include SW in Hz, but not in ppm.  use carrier frequency (sf) to compute SW in ppm
        sw_hz = self.get_xml('connjur:sweepwidth')
        sf_mhz = self.get_xml('connjur:spectralFrequency')
        first_hz = self.get_xml('connjur:firstScalePoint')

        # sweep width of each axis in ppm
        sw_ppm = sw_hz / sf_mhz

        # number of points along each axis per ppm (n points means n-1 intervals)
        self.pts_per_ppm = (self.points - 1) / sw_ppm

        # The "first" point along each axis in ppm.  NOTE: This is the smaller value of the ppm range furthest "right".
        # The "last" point along each axis in ppm.   NOTE: This is the largest value of the ppm range furthest "left".
        self.first_pt_ppm = first_hz / sf_mhz
        self.last_pt_ppm = (first_hz + sw_hz) / sf_mhz

    def print_info(self):
        num_indels = self.mask.size
        empty_num = np.count_nonzero(self.mask)
        nonempty_num = num_indels - empty_num
        empty_percent = (empty_num / num_indels) * 100
        nonempty_percent = (nonempty_num / num_indels) * 100

        print(f'empty indels:    {empty_num:12,d} / {empty_percent:5.2f}%')
        print(f'nonempty indels: {nonempty_num:12,d} / {nonempty_percent:5.2f}%')

    def get_xml(self, name, dtype=float):
        """
        Get the value(s) listed in the xml header for the field "name" and convert to dtype.
        """
        return np.fromiter([e.text for e in self.xml.findall(f'.//{name}', self.ns)], dtype=dtype)

    def ppm_in_mask(self, peak_ppm, box_radius=2):
        """
        map the position of the peak in ppm onto the indel coordinates (points) used by the mask

        consider box_radius points along each dimension above and below the (off-grid) peak position.
        this defines a hypercube around the peak position.  the peak position is considered to be in an empty region if
        any of the points in the hypercube are empty in the mask.

        box_radius can be given as single value (used for all dims) or as a list of values (one for each dim)
        """
        try:
            peak_ppm = np.asarray(peak_ppm)
        except ValueError:
            raise ValueError('failed to convert peak position to numpy array')

        if isinstance(box_radius, int):
            num_dims = len(peak_ppm)
            box_radius = np.repeat(box_radius, num_dims)
        try:
            # convert to array of int and only allow conversions to int where values are preserved
            box_radius = np.asarray(box_radius).astype(int, casting='safe')
        except TypeError:
            raise TypeError('failed to convert box_radius into numpy.array of integers')

        if box_radius.shape != peak_ppm.shape:
            raise ValueError('box radius and peak position do not have same dimensions')

        for axis, p in enumerate(peak_ppm):
            if not (self.first_pt_ppm[axis] <= p <= self.last_pt_ppm[axis]):
                raise PeakOutsideOfMaskError(f'peak position is not in the range of the mask: {peak_ppm}')

        # use position of peak relative to the "last" point  to get index location
        peak_pts = np.abs(peak_ppm - self.last_pt_ppm) * self.pts_per_ppm

        # position of peak in points will not be on-grid, so find all neighboring indels
        peak_pts_down = np.floor(peak_pts).astype(int) - (box_radius - 1)
        peak_pts_up = np.ceil(peak_pts).astype(int) + (box_radius - 1)
        peak_boundaries = [list(range(down, up+1)) for down, up in zip(peak_pts_down, peak_pts_up)]

        # check if each neighbor is empty
        for indel in itertools.product(*peak_boundaries):
            if min(indel) < 0:
                # if the box region goes off grid into negative index, then python would treat that as
                # indexing from end of the array - do NOT want that - just skip the indel as outside the mask
                continue
            try:
                if self.mask[indel]:
                    # found a neighbor that is empty
                    return True
            except IndexError:
                # peak position is at the edge and its "neighbor" is outside the mask
                continue
        # none of the neighbors are empty!
        return False
